@@ -37,19 +37,24 @@ impl From<std::ffi::NulError> for Error {
 
 /// BSD authentication session
 pub struct Session {
-    inner: *mut ffi::auth_session_t,
+    ptr: *mut ffi::auth_session_t,
 }
 
 impl Session {
     /// Open a new BSD Authentication session with the default service
     /// (which can be changed later).
-    pub fn new() -> Self {
+    pub fn new() -> Result<Self, Error> {
         // safety: creates an authentication session,
         // and initializes all members with default values
         //
-        // FIXME: check if return value is null?
-        // Can be if allocation fails
-        Self { inner: unsafe { ffi::auth_open() } }
+        // auth_open returns a null pointer on failure
+        let ptr = unsafe { ffi::auth_open() };
+
+        if ptr == std::ptr::null_mut() {
+            Err(Error::NullSession)
+        } else {
+            Ok(Self { ptr })
+        }
     }
 
     /// Create a Session from a raw auth_session_t pointer
@@ -57,7 +62,7 @@ impl Session {
         if ptr == std::ptr::null_mut() {
             Err(Error::NullSession)
         } else {
-            Ok(Self { inner: ptr })
+            Ok(Self { ptr })
         }
     }
 
@@ -65,14 +70,10 @@ impl Session {
     ///
     /// Consumes the Session
     pub fn into_raw(mut self) -> Result<*mut ffi::auth_session_t, Error> {
-        let null_ptr = std::ptr::null_mut();
-        if self.inner == null_ptr {
-            Err(Error::NullSession)
-        } else {
-            let ret_ptr = self.inner;
-            self.inner = null_ptr;
-            Ok(ret_ptr)
-        }
+        self.check_ptr()?;
+        let ret_ptr = self.ptr;
+        self.ptr = std::ptr::null_mut();
+        Ok(ret_ptr)
     }
 
     /// Request a challenge for the session
@@ -82,8 +83,8 @@ impl Session {
     /// Call is not thread-safe
     pub fn auth_challenge(&self) -> Result<String, Error> {
         // safety: auth_challenge performs null check for the session
-        // So, safe to just pass the inner pointer without a check
-        let c_res = unsafe { ffi::auth_challenge(self.inner) };
+        // So, safe to just pass the ptr pointer without a check
+        let c_res = unsafe { ffi::auth_challenge(self.ptr) };
         if c_res == std::ptr::null_mut() {
             Err(Error::Challenge)
         } else {
@@ -98,17 +99,15 @@ impl Session {
 
     /// Close the specified BSD Authentication session
     ///
-    /// Frees the inner pointer to the session
+    /// Frees the ptr pointer to the session
     /// future calls with the Session will all return Error
     /// 
     /// Inner pointer can be reset with calls that open a new session
     ///
     /// Call is not thread-safe
     pub fn auth_close(&self) -> Result<(), Error> {
-        if self.inner == std::ptr::null_mut() {
-            return Err(Error::NullSession);
-        }
-        let res = unsafe { ffi::auth_close(self.inner) };
+        self.check_ptr()?;
+        let res = unsafe { ffi::auth_close(self.ptr) };
         if res == 0 {
             Err(Error::Close)
         } else {
@@ -121,11 +120,8 @@ impl Session {
     ///
     /// Call is not thread-safe
     pub fn auth_getstate(&self) -> Result<i32, Error> {
-        if self.inner == std::ptr::null_mut() {
-            Err(Error::NullSession)
-        } else {
-            Ok(unsafe { ffi::auth_getstate(self.inner) })
-        }
+        self.check_ptr()?;
+        Ok(unsafe { ffi::auth_getstate(self.ptr) })
     }
 
     /// Set/unset the requested environment variables.
@@ -136,54 +132,45 @@ impl Session {
     ///
     /// Call is not thread-safe
     pub fn auth_setenv(&self) -> Result<(), Error> {
-        if self.inner == std::ptr::null_mut() {
-            Err(Error::NullSession)
-        } else {
-            // safety: auth_setenv does not check the validity of the spool
-            // There is no way to check for an error
-            //
-            // auth_session_t is an opaque type, so we can't check either
-            unsafe { ffi::auth_setenv(self.inner) };
-            Ok(())
-        }
+        self.check_ptr()?;
+        // safety: auth_setenv does not check the validity of the spool
+        // There is no way to check for an error
+        //
+        // auth_session_t is an opaque type, so we can't check either
+        unsafe { ffi::auth_setenv(self.ptr) };
+        Ok(())
     }
 
     /// Clear out any of the requested environment variables.
     ///
     /// Call is not thread-safe
     pub fn auth_clrenv(&self) -> Result<(), Error> {
-        if self.inner == std::ptr::null_mut() {
-            Err(Error::NullSession)
-        } else {
-            // safety: auth_setenv does not check the validity of the spool
-            // There is no way to check for an error
-            //
-            // auth_session_t is an opaque type, so we can't check either
-            unsafe { ffi::auth_clrenv(self.inner) };
-            Ok(())
-        }
+        self.check_ptr()?;
+        // safety: auth_setenv does not check the validity of the spool
+        // There is no way to check for an error
+        //
+        // auth_session_t is an opaque type, so we can't check either
+        unsafe { ffi::auth_clrenv(self.ptr) };
+        Ok(())
     }
 
     /// Get the item value
     ///
     /// Call is not thread-safe
     pub fn auth_getitem(&self, item: AuthItem) -> Result<String, Error> {
-        if self.inner == std::ptr::null_mut() {
-            Err(Error::NullSession)
+        self.check_ptr()?;
+        // safety: auth_getitem also checks for null, so the call is safe
+        let c_res = unsafe { ffi::auth_getitem(self.ptr, item as u32) };
+        if c_res == std::ptr::null_mut() {
+            // return error if the value is unset
+            Err(Error::GetItem)
         } else {
-            // safety: auth_getitem also checks for null, so the call is safe
-            let c_res = unsafe { ffi::auth_getitem(self.inner, item as u32) };
-            if c_res == std::ptr::null_mut() {
-                // return error if the value is unset
-                Err(Error::GetItem)
-            } else {
-                // safety: at this point, the value returned should be a
-                // pointer to a valid UTF-8 C string
-                let c_res = unsafe { CString::from_raw(c_res) };
-                let res = c_res.to_str()?.into();
-                let _ = c_res.into_raw();
-                Ok(res)
-            }
+            // safety: at this point, the value returned should be a
+            // pointer to a valid UTF-8 C string
+            let c_res = unsafe { CString::from_raw(c_res) };
+            let res = c_res.to_str()?.into();
+            let _ = c_res.into_raw();
+            Ok(res)
         }
     }
 
@@ -193,13 +180,14 @@ impl Session {
     ///
     /// Call is not thread-safe
     pub fn auth_setitem(&self, item: AuthItem, value: &str) -> Result<(), Error> {
+        self.check_ptr()?;
         let c_str = CString::new(value)?;
         // safety: auth_setitem checks for null, and sets errno to EINVAL
         // if the auth_session_t* or members are null (so let it do the checks).
         let c_res = match item {
             AuthItem::All => unsafe {
                 // pass null to clear all member items, non-null errors
-                ffi::auth_setitem(self.inner, item as u32, std::ptr::null_mut())
+                ffi::auth_setitem(self.ptr, item as u32, std::ptr::null_mut())
             }
             AuthItem::Interactive => unsafe {
                 let ptr = if value.len() == 0 {
@@ -209,10 +197,10 @@ impl Session {
                     // set to non-null to set the interactive flag
                     c_str.into_raw()
                 };
-                ffi::auth_setitem(self.inner, item as u32, ptr)
+                ffi::auth_setitem(self.ptr, item as u32, ptr)
             }
             _ => unsafe {
-                ffi::auth_setitem(self.inner, item as u32, c_str.into_raw())
+                ffi::auth_setitem(self.ptr, item as u32, c_str.into_raw())
             }
         };
             
@@ -233,19 +221,16 @@ impl Session {
     ///
     /// Call is not thread-safe
     pub fn auth_setoption(&self, name: &str, value: &str) -> Result<(), Error> {
-        if self.inner == std::ptr::null_mut() {
-            Err(Error::NullSession)
-        } else {
-            let n = CString::new(name)?;
-            let v = CString::new(value)?;
-            // safety: auth_setoption checks for null arguments, and argument validity
-            let c_res = unsafe { ffi::auth_setoption(self.inner, n.into_raw(), v.into_raw()) };
+        self.check_ptr()?;
+        let n = CString::new(name)?;
+        let v = CString::new(value)?;
+        // safety: auth_setoption checks for null arguments, and argument validity
+        let c_res = unsafe { ffi::auth_setoption(self.ptr, n.into_raw(), v.into_raw()) };
 
-            if c_res == -1 {
-                Err(Error::SetOption)
-            } else {
-                Ok(())
-            }
+        if c_res == -1 {
+            Err(Error::SetOption)
+        } else {
+            Ok(())
         }
     }
 
@@ -253,29 +238,23 @@ impl Session {
     ///
     /// Call is not thread-safe
     pub fn auth_clroptions(&self) -> Result<(), Error> {
-        if self.inner == std::ptr::null_mut() {
-            Err(Error::NullSession)
-        } else {
-            // safety: auth_clroptions checks for null
-            // in the optlist member, call is safe with the call above
-            unsafe { ffi::auth_clroptions(self.inner) };
-            Ok(())
-        }
+        self.check_ptr()?;
+        // safety: auth_clroptions checks for null
+        // in the optlist member, call is safe with the call above
+        unsafe { ffi::auth_clroptions(self.ptr) };
+        Ok(())
     }
 
     /// Clear the option matching the specified name
     ///
     /// Call is not thread-safe
     pub fn auth_clroption(&self, option: &str) -> Result<(), Error> {
-        if self.inner == std::ptr::null_mut() {
-            Err(Error::NullSession)
-        } else {
-            let opt = CString::new(option)?;
-            // safety: auth_clroption checks for null
-            // in the optlist member, call is safe with the call above
-            unsafe { ffi::auth_clroption(self.inner, opt.into_raw()) };
-            Ok(())
-        }
+        self.check_ptr()?;
+        let opt = CString::new(option)?;
+        // safety: auth_clroption checks for null
+        // in the optlist member, call is safe with the call above
+        unsafe { ffi::auth_clroption(self.ptr, opt.into_raw()) };
+        Ok(())
     }
 
     /// Set BSD Authentication session data to be read into the spool.
@@ -285,23 +264,34 @@ impl Session {
     ///
     /// Call is not thread-safe
     pub fn auth_setdata(&self, data: &mut [u8]) -> Result<(), Error> {
-        if self.inner == std::ptr::null_mut() {
-            Err(Error::NullSession)
-        } else {
-            let d = data.as_mut_ptr() as *mut _;
-            let len = data.len() as u64;
-            // safety: auth_setdata checks for nulls of members. With the null
-            // check for the session above, the call is safe.
-            let c_res = unsafe { ffi::auth_setdata(self.inner, d, len) };
+        self.check_ptr()?;
+        let d = data.as_mut_ptr() as *mut _;
+        let len = data.len() as u64;
+        // safety: auth_setdata checks for nulls of members. With the null
+        // check for the session above, the call is safe.
+        let c_res = unsafe { ffi::auth_setdata(self.ptr, d, len) };
 
-            if c_res == -1 {
-                Err(Error::SetData)
-            } else {
-                Ok(())
-            }
+        if c_res == -1 {
+            Err(Error::SetData)
+        } else {
+            Ok(())
         }
     }
 
+    /// Single function interface to a BSD Authentication session
+    ///
+    /// Functions similarly to a auth_userokay, but does not close the session.
+    ///
+    /// Example:
+    ///
+    /// ```rust
+    /// # use bsd_auth::Session;
+    /// let name = "nobody".to_string();
+    /// let mut passwd = "some_passwd".to_string();
+    ///
+    /// let _session = Session::auth_usercheck(name.as_str(), None, None, Some(&mut passwd)).unwrap();
+    /// ```
+    ///
     /// From `man 3 auth_approval`:
     ///
     /// The auth_usercheck() function operates the same as the auth_userokay()
@@ -334,8 +324,7 @@ impl Session {
         let passwd_ptr = match password {
             Some(p) => {
                 let ptr = CString::new(&*p)?.into_raw();
-                // safety: password guaranteed non-null, and points to valid
-                // memory
+                // safety: pointer is guaranteed non-null, and points to valid memory
                 unsafe { libc::explicit_bzero(p.as_mut_ptr() as *mut _, p.len()) };
                 ptr
             }
@@ -351,24 +340,45 @@ impl Session {
         Self::from_raw(ses_ptr)
     }
     
-    ///  From `man 3 auth_approval`:
+    /// Single function call interface for a BSD Authentication session
     ///
-    ///  Provides a single function call interface.
+    /// Provide a name, and optional style, type and password.
     ///
-    ///  Provided with a user's name in name, and an optional style, type, and password, the auth_userokay() function returns a simple yes/no response.
+    /// If style or type are not provided, the default values will be used.
     ///
-    ///  A return value of true implies failure; a false return value implies success.
-    ///  Other error conditions result in Error.
+    /// Supplying a password uses the non-interactive version of the authentication.
+    /// Not supplying a password uses an interactive authentication mode.
     ///
-    ///  If style is not NULL, it specifies the desired style of authentication to be used.
-    ///  If it is NULL then the default style for the user is used.
-    ///  In this case, name may include the desired style by appending it to the user's name with a single colon (`:') as a separator.
-    ///  If type is not NULL then it is used as the authentication type (such as "auth-myservice").
-    ///  If password is NULL then auth_userokay() operates in an interactive mode with the user on standard input, output, and error.
-    ///  If password is specified, auth_userokay() operates in a non-interactive mode and only tests the specified passwords.
-    ///  This non-interactive method does not work with challenge-response authentication styles.
+    /// Example:
     ///
-    ///  For security reasons, when a password is specified, auth_userokay() will zero out its value before it returns. 
+    /// ```rust
+    /// # use bsd_auth::Session;
+    /// let name = "nobody".to_string();
+    /// let mut passwd = "some_passwd".to_string();
+    ///
+    /// assert!(!Session::auth_userokay(name.as_str(), None, None, Some(&mut passwd)).unwrap());
+    /// ```
+    ///
+    /// From `man 3 auth_approval`:
+    /// 
+    /// ```no_build
+    /// Provides a single function call interface.
+    ///
+    /// Provided with a user's name in name, and an optional style, type, and password, the auth_userokay() function returns a simple yes/no response.
+    ///
+    /// A return value of true implies failure; a false return value implies success.
+    /// Other error conditions result in Error.
+    ///
+    /// If style is not NULL, it specifies the desired style of authentication to be used.
+    /// If it is NULL then the default style for the user is used.
+    /// In this case, name may include the desired style by appending it to the user's name with a single colon (`:') as a separator.
+    /// If type is not NULL then it is used as the authentication type (such as "auth-myservice").
+    /// If password is NULL then auth_userokay() operates in an interactive mode with the user on standard input, output, and error.
+    /// If password is specified, auth_userokay() operates in a non-interactive mode and only tests the specified passwords.
+    /// This non-interactive method does not work with challenge-response authentication styles.
+    ///
+    /// For security reasons, when a password is specified, auth_userokay() will zero out its value before it returns. 
+    /// ```
     ///
     /// For more details see `man 3 auth_approval`
     pub fn auth_userokay(
@@ -410,14 +420,9 @@ impl Session {
     }
     
     /// Get an authentication challenge for the user, with optional style and type
-    ///
-    /// IMPORTANT:
-    ///
-    /// The FFI call returns a session pointer, which is owned by the C library.
-    /// The caller must release ownership of the pointer to prevent a double-free
     /// Example:
     ///
-    /// ```rust,no_build
+    /// ```rust
     /// # use bsd_auth::Session;
     /// /* Create the session and get the challenge */
     /// let (session, _chal) = Session::auth_userchallenge("nobody", Some("passwd"), Some("auth_doas")).unwrap();
@@ -425,9 +430,6 @@ impl Session {
     /// /* Prompt the user for a response */
     /// let mut response = String::from_utf8([1; 1024].to_vec()).unwrap();
     /// session.auth_userresponse(&mut response, 0).unwrap();
-    ///
-    /// /* Release ownership of the inner pointer */
-    /// let _ = session.into_raw();
     /// ```
     ///
     /// From `man 3 auth_approval`:
@@ -489,8 +491,27 @@ impl Session {
         Ok((Self::from_raw(ses_ptr)?, challenge))
     }
     
+    /// Provide a user response for a BSD Authentication session
+    ///
+    /// Consumes the Session due to the FFI call closing the session
+    ///
+    /// Example:
+    ///
+    /// ```rust
+    /// # use bsd_auth::Session;
+    /// let name = "nobody".to_string();
+    /// let style = Some("passwd");
+    /// let mut passwd = "some_passwd".to_string();
+    ///
+    /// let session = Session::auth_usercheck(name.as_str(), style, None, Some(&mut passwd)).unwrap();
+    ///
+    /// let mut res = String::from_utf8([1u8; 1024].to_vec()).unwrap();
+    /// assert!(session.auth_userresponse(&mut res, 0).is_ok());
+    /// ```
+    ///
     /// From `man 3 auth_approval`:
     ///
+    ///```no_build
     /// In addition to the password, the pointer returned by auth_userchallenge()
     /// should be passed in as as and the value of more should be non-zero if the
     /// program wishes to allow more attempts.
@@ -500,27 +521,40 @@ impl Session {
     /// The auth_userresponse() function closes the BSD Authentication session and has the same return value as auth_userokay().
     ///
     /// For security reasons, when a response is specified, auth_userresponse() will zero out its value before it returns.
+    /// ```
     pub fn auth_userresponse(
-        &self,
+        self,
         response: &mut str,
         more: i32,
     ) -> Result<bool, Error> {
+        self.check_ptr()?;
+
         let res_ptr = CString::new(&*response)?.into_raw();
 
         // safety: auth_userresponse checks arguments for null, and clears the
         // memory pointed to by the response pointer
-        let res = unsafe { ffi::auth_userresponse(self.inner, res_ptr, more) };
+        let res = unsafe { ffi::auth_userresponse(self.ptr, res_ptr, more) };
+
+        let _ = self.into_raw()?;
 
         Ok(res != 0)
     }
+
+    fn check_ptr(&self) -> Result<(), Error> {
+        if self.ptr == std::ptr::null_mut() {
+            Err(Error::NullSession)
+        } else {
+            Ok(())
+        }
+    }
 }
 
 impl Drop for Session {
     fn drop(&mut self) {
-        if self.inner != std::ptr::null_mut() {
+        if self.ptr != std::ptr::null_mut() {
             // safety: auth_clean performs null checks
-            // on inner members before freeing
-            unsafe { ffi::auth_close(self.inner); } 
+            // on ptr members before freeing
+            unsafe { ffi::auth_close(self.ptr); } 
         }
     }
 }
@@ -531,7 +565,7 @@ mod tests {
 
     #[test]
     fn test_session() {
-        let _session = Session::new();
+        assert!(Session::new().is_ok());
     }
 
     #[test]
@@ -555,4 +589,14 @@ mod tests {
             assert_eq!(session.auth_getitem(AuthItem::Name).unwrap(), name);
         }
     }
+
+    #[test]
+    fn test_authuserresponse() {
+        let name = "nobody".to_string();
+        let style = Some("passwd");
+        let mut passwd = "some_passwd".to_string();
+        let session = Session::auth_usercheck(name.as_str(), style, None, Some(&mut passwd)).unwrap();
+        let mut res = String::from_utf8([1u8; 1024].to_vec()).unwrap();
+        assert!(session.auth_userresponse(&mut res, 0).is_ok());
+    }
 }
